# use diagnostics ;
# use strict ;
use warnings ;
use sysinfo ; 

our ( $Chatter_limit, $Cron_state, $Database, @Disks, $Failure_repetitions, $Failure_retry_period, $Failure_threshold,
    $Last_cron, $Max_cron_gap, $Minimum_period_between_starts, $Pf, @Pf_restart, $Start_period, $Time_to_die,
    $Umask, %Crontab, %Defines, %Env, %Limit, %Min_free, %Min_inodes, $Min_vmfree, %Original, %Run, @Db, @Processes, @Shutdown_order,
    @Shutdown_tasks, @Startup_tasks,
) ;


sub release { 
    open (RELEASE, "$ENV{ANTELOPE}/bin/epoch -V 2>&1 |" ) ;
    my $release ;
    while ( $release = <RELEASE> ) {
        if ( $release =~ /\*/ ) {
            chomp($release) ;
            $release =~ s/\s*\*\s*//g ;   
            last ;
        }
    } 
    return $release ; 
}

sub email_addresses { 
    my ( $pf, @names ) = @_ ; 
    my (@email, %email, $email) ; 
    foreach $name (@names) { 
	my $ref = pfget ( $pf, $name ) ; 
	if ( ref($ref) ) {
	    if ( ref($ref) eq "HASH" ) { 
		my %hash = %$ref ; 
		@email = keys %hash ; 
	    } elsif ( ref($ref) eq "ARRAY" ) { 
		@email = @$ref ; 
	    } else { 
		rtlog( "bad email value for $name: $$ref" ) ; 
	    }
	} elsif ( defined $ref ) { 
	    $ref =~ s/,/ /g ;
	    @email = split ( " ", $ref ) ;
	} else { 
	    @email = () ; 
	    %email = () ; 
	}
	foreach $email ( @email ) { 
	    $email{$email} = 1; 
	}
    }
    @email = sort keys %email ;
    $email = join(",", @email ) ; 
    return $email ; 
}

sub timetag {
    my($time) = @_ ; 
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) ;
    if ( defined $time ) {
	($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) 
	    = gmtime($time) ;
    } else {
	($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) 
	    = gmtime(time()) ;
    }
    return sprintf ( "%4d-%03d %02d:%02d:%02d", 
	$year+1900, $yday+1, $hour, $min, $sec) ; 
}

sub rtlog {
    my($msg) ;
    my($tag) = &timetag() ; 
    my ($i, $line, @lines ) ; 
    my $l = length($tag) ;
    my $bl = " " x $l ; 
    foreach $_ ( @_ ) {
	$msg = $_ ;
	if ( $msg =~ /^\n/ ) {
	    printf STDERR "\n" ;
	    $msg =~ s/^\n// ; 
	}
	$msg =~ s/\n(.)/\n$bl   $1/g ; 
	printf STDERR "%s %s\n", $tag, $msg ; 
    }
}

sub rtcronlog {
    my($msg) ;
    my($tag) = &timetag() ; 
    my ($i, $line, @lines ) ; 
    my $l = length($tag) ;
    my $bl = " " x $l ; 
    foreach $_ ( @_ ) {
	$msg = $_ ;
	if ( $msg =~ /^\n/ ) {
	    printf STDERR "\n" ;
	    $msg =~ s/^\n// ; 
	}
	$msg =~ s/\n(.)/\n$bl   $1/g ; 
	printf CRONLOG "%s %s\n", $tag, $msg ; 
    }
}

sub init_database {
    if ( ! -f $Database ) {
	open (DB, ">$Database") || 
	    die ( "Can't open $Database : $!\n" ) ; 
	print DB "rtexec1.0\n\n" ; 
	close DB ; 
    }
    @Db = dbopen ( $Database, "r+" ) ; 
    @Db = dblookup (@Db, "", "process", "", "" ) ; 
}

sub dblog_shutdown {
    my ( $who, $why ) = @_ ; 
    if ( ! defined @Db ) { 
	init_database() ; 
    }
    my(@db) = dblookup ( @Db, "", "shutdown", "", "" ) ; 
    my ( $now ) = time() ;
    eval { 
	$db[3] = dbaddnull(@db) ;
	    dbputv ( @db,
	    "down", $now, 
	    "who", $who,
	    "why", $why ) ;
    } ;
    if ( $@ ne "" ) {
	&rtlog("$@") ;
    }
}

sub dblog {
    my ($name, $execution, $pid) = @_ ; 
    if ( ! defined @Db ) { 
	init_database() ; 
    }
    my(@db) = dblookup ( @Db, "", "process", "name", $name ) ; 
    my ($filename)  ;
    $pid = -1 if ! defined $pid ;
    my ( $now ) = time() ;
    if ( $db[3] < 0 ) {
	eval { 
	    $db[3] = dbaddnull (@db) ;
	    dbputv ( @db,
		"name", $name,
		"pid", $pid,
		"first_start", $now,
		"totstarts", 1,
		"starts", 1, 
		"last_start", $now, 
		"execution", $execution ) ; 
	} ;
	if ( $@ ne "" ) {
	    &rtlog("$@") ;
	}

    } else { 
	my ($totstarts, $starts) = dbgetv(@db, qw(totstarts starts)) ;
	&dbputv ( @db,
		"pid", $pid,
		"totstarts", $totstarts+1,
		"starts", $starts+1,
		"last_start", $now, 
		"execution", $execution ) ; 
    }
    $filename = dbquery ( @db, "dbTABLE_FILENAME" ) ; 
# rtm looks at the time on this table to decide when a new process has started
    utime $now, $now, $filename ; 

    &history($name, $execution, $pid) ; 
}

sub cronlog {
    my ($name, $execution, $pid) = @_ ; 
    if ( ! defined @Db ) { 
	init_database() ; 
    }
    $name =~ s/^cron: // ;
    my(@db) = dblookup ( @Db, "", "cron", "name", $name ) ; 
    my ($filename)  ;
    my ( $now ) = time() ;
    my (@db2) = dblookup ( @db, "", "", "", "dbNULL" ) ; 
    my ($null) = dbgetv (@db2, "last_end")  ;
    if ( $db[3] < 0 ) {
	eval { 
	    $db[3] = dbaddnull (@db) ;

	} ;
	if ( $@ ne "" ) {
	    &rtlog("$@") ;
	    return ;
	}

    } 
    $pid = -1 if ! defined $pid ;
    dbputv ( @db,
	"name", $name,
	"pid", $pid,
	"last_start", $now, 
	"last_end", $null,
	"result", 9999, 
	"execution", $execution ) ; 
    $filename = dbquery ( @db, "dbTABLE_FILENAME" ) ; 
# rtm looks at the time on this table to decide when a new cron has started
    utime $now, $now, $filename ; 

    &cron_history($name, $execution, $pid) ; 
}

sub endcron { 
    my ($pid, $name, $status ) = @_ ;  
    if ( ! defined @Db ) { 
	init_database() ; 
    }
    $name =~ s/^cron: // ;
    my(@db) = dblookup ( @Db, "", "cron", "name", $name ) ; 
    my ($filename)  ;
    my ( $now ) = time() ;
    if ( $db[3] < 0 ) {
	&rtlog("No previous cron entry for $id: $name : $status !!") ;
	return ;
    } else { 
	$dbpid = dbgetv ( @db, "pid" ) ;
	if ( $dbpid != $pid ) {
	    &rtlog("previous cron entry for $name is $dbpid, not $pid!!") ;
	}
	dbputv ( @db, "pid", $pid, "last_end", $now, "result", $status ) ; 
    }
    $filename = dbquery ( @db, "dbTABLE_FILENAME" ) ; 
# rtm looks at the time on this table to decide when a new cron has started
    utime $now, $now, $filename ; 
}

sub history {
    my ($name, $execution, $pid ) = @_ ; 
    if ( ! defined @Db ) { 
	init_database() ; 
    }
    $pid = -1 if ! defined $pid ;
    my @db = dblookup ( @Db, "", "history", "", "" ) ; 
    eval { 
	$db[3] = dbaddnull(@db) ;
	  dbputv ( @db,
		"name", $name,
		"start", time(),
		"pid", $pid,
		"execution", $execution ) ; 
    } ;
    if ( $@ ) {
	&rtlog("$@: couldn't add new record to $Database.history") ;
    }
}

sub cron_history {
    my ($name, $execution, $pid ) = @_ ; 
    if ( ! defined @Db ) { 
	init_database() ; 
    }
    $pid = -1 if ! defined $pid ;
    my @db = dblookup ( @Db, "", "cron_history", "", "" ) ; 
    eval { 
	$db[3] = dbaddnull(@db) ;
	  dbputv ( @db,
		"name", $name,
		"start", time(),
		"pid", $pid,
		"execution", $execution ) ; 
    } ;
    if ( $@ ) {
	&rtlog("$@: couldn't add new record to $Database.history") ;
    }
}

sub setlimits {
    my ($resource) ; 
    foreach $resource (keys %Limit) {
	&limit($resource, $Limit{$resource}) ; 
    }
}

sub clrenv {
    my ($key, $msg) ;
    %Original = %ENV ;
    foreach $key ( keys %ENV ) {
	if ( $opt_d ) {
	    $msg = sprintf ( "deleting ENV($key) = '%s'", $ENV{$key} ) ; 
	    print STDERR "$msg\n" ;
	}
	delete $ENV{$key};
    }
}

sub substituteDefines { 
    my ($line) = @_ ;
    die "Defines not defined\n" if ! defined %Defines ;
    # $line =~ s/\$(\w+)\b/$Defines{$1}||$&/ge ; 
    $line =~ s/\$(\w+)\b/defined $Defines{$1}?$Defines{$1}:$&/ge ;
    $line =~ s/\$\{(\w+)\}/defined $Defines{$1}?$Defines{$1}:$&/ge ;
    return $line ;
}

sub setenv {
    my ($key, $value, $msg, $r) ;
    die "Original not defined\n" if ! %Original ;

    foreach $key ( keys %Env ) {
	$value = $Env{$key} ;
	$value = substituteDefines($value) ; 

	if ( $value =~ /\$Original\{(\w+)\}/ ) { 
	    $r = $Original{$1} ;
	    $value =~ s/(\$Original\{\w+\})/$r/g ;
	    print STDERR "Please switch the Env element $key to use PRESERVE rather than \$Original\n";
	}
	if ( $value =~ /^PRESERVE\s*\|\|\s*(.*)/ ) { 
	    $value = $Original{$key} ? $Original{$key} : $1 ; 
	} elsif ( $Env{$key} =~ /^PRESERVE/ ) { 
	    if (defined $Original{$key}) {
		$value = $Original{$key} 
	    } else { 
		$value = undef ; 
	    }
	}

	if ( defined $value ) { 
	    print STDERR "setting ENV{$key} = $value\n" if $opt_d ;
	    $ENV{$key} = $value ;
	} else { 
	    print STDERR "deleting $key from environment\n" if $opt_d ;
	    delete $ENV{$key} ;
	}
    }
}

sub cronitemcmp {
    my ( $t, $c ) = @_ ;
    return 1 if $c eq "*" ;
    my @c = split(',', $c) ; 
    my $n = grep($t == $_, @c) ; 
    return ($n > 0) ;
}

sub croncmp { 
    my ( $min,  $hour,  $month,  $day,  $dow, 
	 $cmin, $chour, $cmonth, $cday, $cdow ) = @_ ;

    my ( $result ) ; 

    if ( cronitemcmp($min, $cmin)
	&& cronitemcmp($hour, $chour) ) {
	if ($cdow eq "*") {
	    if (cronitemcmp($month, $cmonth) && cronitemcmp($day, $cday)) {
		# month/day match 
		$result = 1 ; 
	    } else { 
		# month/day don't match && dow is '*' 
		$result = 0 ; 
	    }
	} elsif ( cronitemcmp($dow, $cdow) ) { 
	    # day of week matches 
	    $result = 1 ; 
	} elsif ( ($cmonth  ne "*" || $cday ne "*")
	      && (cronitemcmp($month, $cmonth) && cronitemcmp($day, $cday)) ) { 
	    # day of week doesn't match, 
	    #	but month/day is specified and matches
	    $result = 1 ; 
        } else {
	    # dow doesn't match, month/day is not specified 
	    $result = 0 ; 
	}
    } else { 
	$result = 0 ; 
    }
    return $result ; 
}

sub crontab {
    my ($time) = @_ ;

    my ($usec,$umin,$uhour,$uday,$umonth,$uyear,$udow,$uyday,$uisdst) =
					     gmtime($time);
    my ( $result ) ; 
    $umonth++ ;
    $uyear += 1900 ;

    my ($lsec,$lmin,$lhour,$lday,$lmonth,$lyear,$ldow,$lyday,$lisdst) =
					     localtime($time);
    $lmonth++ ;
    $lyear += 1900 ;

    my ( $utc, $cmin, $chour, $cday, $cmonth, $cdow, $cmd, $task ) ; 
    foreach $task ( keys %Crontab ) { 
	( $utc, $cmin, $chour, $cday, $cmonth, $cdow, $cmd ) = 
	    split ( ' ', $Crontab{$task}, 7 ) ;

	if ( $utc eq "LOCAL" ) { 
	    $result = croncmp($lmin, $lhour, $lmonth, $lday, $ldow, 
			      $cmin, $chour, $cmonth, $cday, $cdow ) ; 
	} else { 
	    if ( $utc ne "UTC" ) { 
		rtlog ( "timezone code '$utc' is neither 'LOCAL' or 'UTC': using 'UTC'" ) ; 
	    }
	    $result = croncmp($umin, $uhour, $umonth, $uday, $udow, 
			      $cmin, $chour, $cmonth, $cday, $cdow ) ; 
	}
	if ( $result ) { 
	    &rtcronlog("cron: $task : $cmd") ;
	    my $pid = &spawn ( "cron: $task", $cmd, $time ) ;
	    # &cron_history($task, $cmd, $pid); 
	}
    }
}

sub run_cronjobs { 
    my ($time) = time(); 
    my ($counter) = 0 ; 
    if ( $Last_cron < $time - $Max_cron_gap ) { 
	rtlog(sprintf("** system clock appears to have jumped forward from\n\t%s\nto\t%s", 
		strtime($Last_cron),
		strtime($time))) ; 
	$Last_cron = $time ;
    } elsif ( $Last_cron - $time > 61 ) { 
	rtlog(sprintf("** system clock appears to have jumped backward from\n\t%s\nto\t%s", 
		strtime($Last_cron),
		strtime($time))) ; 
	$Last_cron = $time ; 
	rtlog ( "resetting last cron time; some duplication of cron jobs may occur" ) ; 
    }

    $Max_cron_gap = pfget($Pf, "max_cron_gap" ) ;
    for ( ; $Last_cron <= $time ; $Last_cron += 60 ) { 
	&crontab($Last_cron) ; 
	utime $Last_cron, $Last_cron, $Cron_state ; 
	last if ($counter++ > 10) ;
    }
}

sub croncvtdash { 
    my ($t) = @_ ; 
    if ( $t =~ /-/ ) { 
	my @old = split(',', $t) ; 
	my ($i, $j, $n, $start, $stop, @new) ;
	$n = @old ;
	@new = () ;
	for($i=0 ; $i<$n ; $i++) { 
	    if ( $old[$i] =~ /(\d+)-(\d+)/ ) {
		$start = $1 ; 
		$stop = $2 ; 
		for ( $j=$start ; $j<=$stop ; $j++ ) { 
		    push (@new, $j) ; 
		}
	    } else { 
		push ( @new, $old[$i] ) ;
	    }
	}
	$t = join(',', @new) ; 
    }
    return $t ;
}

sub get_crontab {
    $Max_cron_gap = pfget($Pf, "max_cron_gap" ) ;
    my $ref = pfget ( $Pf, 'crontab' ) ; 
    %Crontab = %$ref ;
    my ( $utc, $cmin, $chour, $cday, $cmonth, $cdow, $cmd, $task ) ; 
    foreach $task ( keys %Crontab ) { 
	( $utc, $cmin, $chour, $cday, $cmonth, $cdow, $cmd ) = 
	    split ( ' ', $Crontab{$task}, 7 ) ;
	$cmin   = croncvtdash($cmin) ;
	$chour  = croncvtdash($chour) ;
	$cday   = croncvtdash($cday) ;
	$cmonth = croncvtdash($cmonth) ;
	$cdow   = croncvtdash($cdow) ;
	$Crontab{$task} = join(' ', 
	    ( $utc, $cmin, $chour, $cday, $cmonth, $cdow, $cmd ) ) ;
    }
}

sub get_cron_state { 

    $Cron_state = "state/cron" ;
    if ( ! -e $Cron_state ) { 
	system ( "touch $Cron_state" ) ; 
    }
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	    $atime,$mtime,$ctime,$blksize,$blocks)
				    = stat($Cron_state);
    $Last_cron = $mtime ;
    # printf STDERR "last cron was run at %s\n", strtime($Last_cron) ; 
    $Last_cron += 60 ;
    if ( $Last_cron < time() - $Max_cron_gap ) { 
	printf STDERR "** cron clock starts now, not when cron last ran at %s\n", 
		strtime($Last_cron) ; 
	$Last_cron = time() ;
    }

    $Last_cron = int ($Last_cron / 60) * 60 ;
    # printf STDERR "next cron will run for %s\n", strtime($Last_cron) ; 
}

sub get_Defines { 
    my ( $pf ) = @_ ;
    my ( $ref, $key, $value ) ; 
    my $ANTELOPE = pfget ( $pf, 'ANTELOPE' ) ; 
    my $ROOT = pfget ( $pf, 'ROOT' ) ; 
    $ref = pfget ($pf, 'Defines') ;
    %Defines = %$ref ;

    %Original = %ENV if ! %Original ;

    foreach $key ( keys %Defines ) {
	$value = $Defines{$key} ;
	if ( $value =~ /^PRESERVE\s*\|\|\s*(.*)/ ) { 
	    $value = $Original{$key} ? $Original{$key} : $1 ; 
	} elsif ( $value =~ /^PRESERVE/ ) { 
	    $value = $Original{$key} ;
	} 
	$value = substituteDefines($value) ; 
	$Defines{$key} = $value ;
    }

#  for the benefit of people running older versions of rtexec
    if ( defined $ANTELOPE) { 
	$Defines{'ANTELOPE'} = $ANTELOPE ;
	print STDERR "Please move the ANTELOPE definition to the Defines array\n" ;
    }
    if ( defined $ROOT) { 
	$Defines{'ROOT'} = $ROOT ;
	print STDERR "Please move the ROOT definition to the Defines array\n" ;
    }
}

sub eval_pf {
    my ( $pf ) = @_ ; 
    my ( $ref, $key ) ; 

    get_Defines ($pf) ;

    $ref = pfget ( $pf, 'Env' ) ; 
    %Env = %$ref ;

    $Umask = pfget($pf, "umask" ) ; 

    $ref = pfget ( $pf, "Shutdown_when_task_dies" ) ; 
    if ( defined $ref ) { 
	%Shutdown = %$ref ;
    } else { 
	%Shutdown = () ; 
    }

    $Start_period = pfget ( $pf, 'Start_period' ) ; 
    $Minimum_period_between_starts = pfget ( $pf, 'Minimum_period_between_starts' ) ; 

    $ref = pfget($pf, 'Processes' ) ; 
    @Processes = @$ref ;

    $ref = pfget ( $pf, 'Run' ) ; 
    %Run = %$ref ;
    foreach $key ( keys %Run ) { 
	$ref = yesno($Run{$key}) ; 
	if ( $ref == -1 ) { 
	    $Run{$key} = 1 ; 
	} elsif ($ref == 0) { 
	    $Run{$key} = 0 ; 
	} else { 
	    $Run{$key} = 0 ; 
	    &rtlog ("run value '$Run{$key}' for $key is not recognized.") ;
	}
    }

    $ref = pfget($pf, 'Pf_restart' ) ; 
    @Pf_restart = @$ref ;

    $ref = pfget($pf, 'disks' ) ; 
    @Disks = @$ref ;

    $Chatter_limit = pfget ( $pf, 'Chatter_limit' ) ;
    $Min_vmfree = pfget ( $pf, 'Min_vmfree' ) ;


    $ref = pfget($pf, 'Shutdown_order' ) ; 
    @Shutdown_order = @$ref ;

    $ref = pfget($pf, 'Startup_tasks' ) ; 
    @Startup_tasks = @$ref ;

    $ref = pfget($pf, 'Shutdown_tasks' ) ; 
    @Shutdown_tasks = @$ref ;

    $Time_to_die = pfget ( $pf, 'Time_to_die' ) ; 

    $ref = pfget ( $pf, 'Limit' ) ; 
    %Limit = %$ref ;

    $Database = pfget ( $pf, 'Database' ) ; 

    $Failure_threshold 	 = pfget($pf, "Failure_threshold" ) ;
    $Failure_repetitions = pfget($pf, "Failure_repetitions" ) ;
    $Failure_retry_period = pfget($pf, "Failure_retry_period" ) ;


    &get_crontab() ;
}

1;
